import logging
from collections import defaultdict

from django.core.exceptions import ObjectDoesNotExist
from django.db import transaction
from django.db.models.manager import Manager
from rest_framework import serializers
from rest_framework.utils import html

from core import models
from core.models import Feedback, TutorSubmissionAssignment
from util.factories import GradyUserFactory

from .generic import DynamicFieldsModelSerializer

log = logging.getLogger(__name__)
user_factory = GradyUserFactory()


class FeedbackCommentDictionarySerializer(serializers.ListSerializer):

    def to_internal_value(self, comment_dict):
        """ Converts a line_no -> comment list dictionary back to a list
        of comments. Currently we do not have any information about the
        feedback since it is not availiable in this scope. Feedback is
        responsible to add it later on update/creation """
        if html.is_html_input(comment_dict):
            comment_dict = html.parse_html_list(comment_dict)

        if not isinstance(comment_dict, dict):
            raise serializers.ValidationError(
                'Comments have to be provided as a dict'
                'with: line -> list of comments'
            )

        ret = []
        errors = []

        for line, comment in comment_dict.items():
            try:
                comment['of_line'] = line
                validated = self.child.run_validation(comment)
            except serializers.ValidationError as err:
                errors.append(err.detail)
            else:
                ret.append(validated)
                errors.append({})

        if any(errors):
            raise serializers.ValidationError(errors)

        return ret

    def to_representation(self, comments):
        """ Provides a dict where all the keys correspond to lines and contain
        a list of comments on that line. """
        if isinstance(comments, Manager):
            comments = comments.all()

        ret = defaultdict(list)
        for comment in comments:
            ret[comment.of_line].append(self.child.to_representation(comment))
        return ret


class FeedbackCommentSerializer(serializers.ModelSerializer):
    of_tutor = serializers.StringRelatedField(source='of_tutor.username')

    class Meta:
        model = models.FeedbackComment
        fields = ('text',
                  'created',
                  'of_tutor',
                  'of_line',
                  'visible_to_student')
        read_only_fields = ('created', 'of_tutor')
        extra_kwargs = {
            'of_feedback': {'write_only': True},
            'of_line': {'write_only': True},
        }
        list_serializer_class = FeedbackCommentDictionarySerializer


class FeedbackSerializer(DynamicFieldsModelSerializer):
    pk = serializers.ReadOnlyField(source='of_submission.pk')
    assignment_pk = serializers.UUIDField(write_only=True)
    feedback_lines = FeedbackCommentSerializer(many=True, required=False)

    @transaction.atomic
    def create(self, validated_data) -> Feedback:
        assignment = validated_data.pop('assignment_pk')
        feedback_lines = validated_data.pop('feedback_lines', [])

        feedback = Feedback.objects.create(of_submission=assignment.submission,
                                           **validated_data)

        for comment in feedback_lines:
            models.FeedbackComment.objects.create(
                of_feedback=feedback,
                of_tutor=self.context['request'].user,
                **comment
            )

        assignment.set_done()
        return Feedback.objects.get(of_submission=assignment.submission)

    @transaction.atomic
    def update(self, instance, validated_data):
        for comment in validated_data.pop('feedback_lines', []):
            models.FeedbackComment.objects.update_or_create(
                of_feedback=instance,
                of_tutor=self.context['request'].user,
                of_line=comment.get('of_line'),
                defaults={'text': comment.get('text')})

        return super().update(instance, validated_data)

    def validate_assignment_pk(self, assignment_pk):
        try:
            assignment = TutorSubmissionAssignment.objects.get(
                pk=assignment_pk)
        except ObjectDoesNotExist as err:
            raise serializers.ValidationError('No assignment for given id.')

        return assignment

    def validate(self, data):
        log.debug("Validate feedback data: %s", data)
        score = data.get('score')
        assignment = data.get('assignment_pk')

        submission = assignment.submission
        if not 0 <= score <= submission.type.full_score:
            raise serializers.ValidationError(
                f'Score has to be in range [0..{submission.type.full_score}].')

        has_full_score = score == submission.type.full_score
        has_feedback_lines = ('feedback_lines' in data and
                              len(data['feedback_lines']) > 0)
        if not has_full_score and not has_feedback_lines:
            raise serializers.ValidationError(
                'Sorry, you have to explain why this does not get full score')

        http_method = self.context['request'].method
        if hasattr(submission, 'feedback') and http_method == 'POST':
            raise serializers.ValidationError(
                'Feedback for this submission already exists')

        for comment in data.get('feedback_lines', {}):
            lines_in_submission = len(submission.text.split('\n'))
            if not 0 < comment['of_line'] <= lines_in_submission:
                raise serializers.ValidationError(
                    "Cannot comment line number %d of %d" % (
                        comment['of_line'], lines_in_submission))

        return data

    class Meta:
        model = Feedback
        fields = ('pk', 'assignment_pk', 'is_final', 'score', 'feedback_lines')
