from django.core.exceptions import ObjectDoesNotExist
from rest_framework import serializers

from core.models import (GeneralTaskSubscription, Submission,
                         TutorSubmissionAssignment)
from core.serializers import DynamicFieldsModelSerializer, FeedbackSerializer


class AssignmentSerializer(DynamicFieldsModelSerializer):
    submission_pk = serializers.ReadOnlyField(
        source='submission.pk')

    class Meta:
        model = TutorSubmissionAssignment
        fields = ('pk', 'submission_pk', 'is_done',)


class SubmissionAssignmentSerializer(DynamicFieldsModelSerializer):
    text = serializers.ReadOnlyField()
    type_pk = serializers.ReadOnlyField(source='type.pk')
    full_score = serializers.ReadOnlyField(source='type.full_score')

    class Meta:
        model = Submission
        fields = ('pk', 'type_pk', 'text', 'full_score')


class AssignmentDetailSerializer(DynamicFieldsModelSerializer):
    submission = SubmissionAssignmentSerializer()
    feedback = FeedbackSerializer(source='submission.feedback')

    class Meta:
        model = TutorSubmissionAssignment
        fields = ('pk', 'feedback', 'submission', 'is_done',)


class SubscriptionSerializer(DynamicFieldsModelSerializer):
    owner = serializers.ReadOnlyField(source='owner.username')
    query_key = serializers.CharField(required=False)
    assignments = AssignmentSerializer(read_only=True, many=True)

    def validate(self, data):
        data['owner'] = self.context['request'].user

        if 'query_key' in data != \
                data['query_type'] == GeneralTaskSubscription.RANDOM:
            raise serializers.ValidationError(
                f'The {data["query_type"]} query_type does not work with the'
                f'provided key')

        try:
            GeneralTaskSubscription.objects.get(
                owner=data['owner'],
                query_type=data['query_type'],
                query_key=data.get('query_key', None))
        except ObjectDoesNotExist:
            pass
        else:
            raise serializers.ValidationError(
                'The user already has the subscription')

        return data

    def create(self, validated_data) -> GeneralTaskSubscription:
        return GeneralTaskSubscription.objects.create(**validated_data)

    class Meta:
        model = GeneralTaskSubscription
        fields = (
            'pk',
            'owner',
            'query_type',
            'query_key',
            'assignments',
            'feedback_stage')
