from rest_framework import status
from rest_framework.test import APIClient, APITestCase

from core import models
from core.models import (GeneralTaskSubscription, Submission, SubmissionType,
                         SubscriptionEnded)
from util.factories import GradyUserFactory, make_test_data


class GeneralTaskSubscriptionRandomTest(APITestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user_factory = GradyUserFactory()

    def setUp(self):
        self.t = self.user_factory.make_tutor()
        self.s1 = self.user_factory.make_student()
        self.s2 = self.user_factory.make_student()

        self.submission_type = SubmissionType.objects.create(
            name='submission_01', full_score=14)
        self.submission_01 = Submission.objects.create(
            type=self.submission_type, student=self.s1.student,
            text='I really failed')
        self.submission_02 = Submission.objects.create(
            type=self.submission_type, student=self.s2.student,
            text='I like apples')

        self.subscription = GeneralTaskSubscription.objects.create(
            owner=self.t, query_type=GeneralTaskSubscription.RANDOM)

    def test_subscription_gets_an_assignment(self):
        self.subscription._create_new_assignment_if_subscription_empty()
        self.assertEqual(1, self.subscription.assignments.count())

    def test_first_work_assignment_was_created_unfinished(self):
        self.subscription._create_new_assignment_if_subscription_empty()
        self.assertFalse(self.subscription.assignments.first().is_done)

    def test_subscription_raises_error_when_depleted(self):
        self.submission_01.delete()
        self.submission_02.delete()
        try:
            self.subscription._create_new_assignment_if_subscription_empty()
        except SubscriptionEnded as err:
            self.assertFalse(False)
        else:
            self.assertTrue(False)

    def test_can_prefetch(self):
        self.subscription._create_new_assignment_if_subscription_empty()
        self.subscription._eagerly_reserve_the_next_assignment()
        self.assertEqual(2, self.subscription.assignments.count())

    def test_oldest_assignment_is_current(self):
        assignment = self.subscription.get_oldest_unfinished_assignment()
        self.assertEqual(assignment,
                         self.subscription.get_oldest_unfinished_assignment())


class TestApiEndpoints(APITestCase):

    @classmethod
    def setUpTestData(cls):
        cls.data = make_test_data(data_dict={
            'submission_types': [
                {
                    'name': '01. Sort this or that',
                    'full_score': 35,
                    'description': 'Very complicated',
                    'solution': 'Trivial!'
                },
                {
                    'name': '02. Merge this or that or maybe even this',
                    'full_score': 35,
                    'description': 'Very complicated',
                    'solution': 'Trivial!'
                },
                {
                    'name': '03. This one exists for the sole purpose to test',
                    'full_score': 30,
                    'description': 'Very complicated',
                    'solution': 'Trivial!'
                }
            ],
            'students': [
                {'username': 'student01'},
                {'username': 'student02'}
            ],
            'tutors': [
                {'username': 'tutor01'},
                {'username': 'tutor02'}
            ],
            'reviewers': [
                {'username': 'reviewer'}
            ],
            'submissions': [
                {
                    'text': 'function blabl\n'
                            '   on multi lines\n'
                            '       for blabla in bla:\n'
                            '           lorem ipsum und so\n',
                    'type': '01. Sort this or that',
                    'user': 'student01',
                    'feedback': {
                        'score': 5,
                        'is_final': True,
                        'feedback_lines': {
                            '1': [{
                                'text': 'This is very bad!',
                                'of_tutor': 'tutor01'
                            }],
                        }

                    }
                },
                {
                    'text': 'function blabl\n'
                            '       asasxasx\n'
                            '           lorem ipsum und so\n',
                    'type': '02. Merge this or that or maybe even this',
                    'user': 'student01'
                },
                {
                    'text': 'function blabl\n'
                            '   on multi lines\n'
                            '       asasxasx\n'
                            '           lorem ipsum und so\n',
                    'type': '03. This one exists for the sole purpose to test',
                    'user': 'student01'
                },
                {
                    'text': 'function lorem ipsum etc\n',
                    'type': '03. This one exists for the sole purpose to test',
                    'user': 'student02'
                },
            ]}
        )

    def test_can_create_a_subscription(self):
        client = APIClient()
        client.force_authenticate(user=self.data['tutors'][0])

        response = client.post('/api/subscription/', {'query_type': 'random'})

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_create_subscription_and_get_one_assignment(self):
        client = APIClient()
        client.force_authenticate(user=self.data['tutors'][0])

        response = client.post('/api/subscription/', {'query_type': 'random'})

        self.assertEqual('tutor01', response.data['owner'])

    def test_subscription_has_next_and_current_assignment(self):
        client = APIClient()
        client.force_authenticate(user=self.data['tutors'][0])

        response_subs = client.post(
            '/api/subscription/', {'query_type': 'random'})
        subscription_id = response_subs.data['pk']
        assignment_pk = response_subs.data['assignments'][0]['pk']

        response = client.get(
            f'/api/subscription/{subscription_id}/assignments/current/')

        self.assertEqual(1, len(response_subs.data['assignments']))
        self.assertEqual(assignment_pk,
                         response.data['pk'])

        response_next = client.get(
            f'/api/subscription/{subscription_id}/assignments/next/')
        response_detail_subs = \
            client.get(f'/api/subscription/{subscription_id}/')

        self.assertEqual(2, len(response_detail_subs.data['assignments']))
        self.assertNotEqual(assignment_pk, response_next.data['pk'])

    def test_subscription_can_assign_to_student(self):
        client = APIClient()
        client.force_authenticate(user=self.data['reviewers'][0])

        student01 = self.data['students'][0]

        response = client.post(
            '/api/subscription/', {
                'query_type': 'student',
                'query_key': student01.student.student_id,
                'stage': 'feedback-creation'
            })

        assignments = response.data['assignments']
        self.assertEqual(2, len(assignments))

    def test_all_stages_of_the_subscription(self):
        client = APIClient()
        client.force_authenticate(user=self.data['tutors'][0])

        # The tutor corrects something
        response = client.post(
            '/api/subscription/', {
                'query_type': 'random',
                'feedback_stage': 'feedback-creation'
            })

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        assignment_pk = response.data['assignments'][0]['pk']
        response = client.post(
            f'/api/feedback/', {
                "score": 23,
                "assignment_pk": assignment_pk,
                "feedback_lines": {
                    2: {"text": "< some string >"},
                    3: {"text": "< some string >"}
                }
            }
        )
        print(response, response.data)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        # some other tutor reviews it
        client.force_authenticate(user=self.data['tutors'][1])

        response = client.post(
            '/api/subscription/', {
                'query_type': 'random',
                'feedback_stage': 'feedback-validation'
            })

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        subscription_id = response.data['pk']
        response = client.get(
            f'/api/subscription/{subscription_id}/assignments/current/')

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        submission_id_in_database = models.Feedback.objects.filter(
            is_final=False).first().of_submission.submission_id
        submission_id_in_response = \
            response.data['submission']['pk']

        self.assertEqual(
            str(submission_id_in_database),
            submission_id_in_response)
