""" All API views that are used to retrieve data from the database. They
can be categorized by the permissions they require. All views require a
user to be authenticated and most are only accessible by one user group """
import logging

from django.conf import settings
from rest_framework import generics, mixins, viewsets
from rest_framework.decorators import api_view
from rest_framework.response import Response

from core import models
from core.models import ExamType, StudentInfo, SubmissionType
from core.permissions import IsReviewer, IsStudent
from core.serializers import (ExamSerializer, StudentInfoSerializer,
                              StudentInfoSerializerForListView,
                              SubmissionSerializer, SubmissionTypeSerializer,
                              TutorSerializer)

log = logging.getLogger(__name__)


@api_view()
def get_jwt_expiration_delta(request):
    return Response({'timeDelta': settings.JWT_AUTH['JWT_EXPIRATION_DELTA']})


@api_view()
def get_user_role(request):
    return Response({'role': request.user.role})


class StudentSelfApiView(generics.RetrieveAPIView):
    """ Gets all data that belongs to one student """
    permission_classes = (IsStudent,)
    serializer_class = StudentInfoSerializer

    def get_object(self) -> StudentInfo:
        """ The object in question is the student associated with the requests
        user. Since the permission IsStudent is satisfied the member exists """
        if self.request.user.is_superuser:
            return StudentInfo.objects.last()
        return self.request.user.student


class StudentSelfSubmissionsApiView(generics.ListAPIView):
    permission_classes = (IsStudent, )
    serializer_class = SubmissionSerializer

    def get_queryset(self):
        return self.request.user.student.submissions


class StudentReviewerApiViewSet(viewsets.ReadOnlyModelViewSet):
    """ Gets a list of all students without individual submissions """
    permission_classes = (IsReviewer,)
    queryset = StudentInfo.objects.all()
    serializer_class = StudentInfoSerializerForListView


class ExamApiViewSet(viewsets.ReadOnlyModelViewSet):
    """ Gets a list of an individual exam by Id if provided """
    permission_classes = (IsReviewer,)
    queryset = ExamType.objects.all()
    serializer_class = ExamSerializer


class TutorApiViewSet(
        mixins.RetrieveModelMixin,
        mixins.CreateModelMixin,
        mixins.DestroyModelMixin,
        mixins.ListModelMixin,
        viewsets.GenericViewSet):
    """ Api endpoint for creating, listing, viewing or deleteing tutors """
    permission_classes = (IsReviewer,)
    queryset = models.UserAccount.objects.filter(role='Tutor')
    serializer_class = TutorSerializer
    lookup_field = 'username'
    lookup_url_kwarg = 'username'


class SubmissionTypeApiView(viewsets.ReadOnlyModelViewSet):
    """ Gets a list or a detail view of a single SubmissionType """
    queryset = SubmissionType.objects.all()
    serializer_class = SubmissionTypeSerializer
