import logging

from rest_framework import mixins, status, viewsets
from rest_framework.response import Response

from core import models, permissions, serializers

log = logging.getLogger(__name__)


class FeedbackApiView(
        mixins.CreateModelMixin,
        mixins.RetrieveModelMixin,
        viewsets.GenericViewSet):
    """ Gets a list of an individual exam by Id if provided """
    permission_classes = (permissions.IsTutorOrReviewer,)
    queryset = models.Feedback.objects.all()
    serializer_class = serializers.FeedbackSerializer
    lookup_field = 'of_submission__pk'
    lookup_url_kwarg = 'submission_pk'

    def _tutor_attempts_to_change_final_feedback(self, serializer):
        feedback_is_final = serializer.instance.is_final
        user_is_tutor = self.request.user.role == models.UserAccount.TUTOR
        return feedback_is_final and user_is_tutor

    def _request_user_does_not_own_assignment(self, serializer):
        assignment = serializer.validated_data['assignment_pk']
        return assignment.subscription.owner != self.request.user

    def _tutor_attempts_to_set_first_feedback_final(self, serializer):
        is_final_set = serializer.validated_data.get('is_final', False)
        user_is_tutor = self.request.user.role == models.UserAccount.TUTOR
        return is_final_set and user_is_tutor

    def _tutor_is_allowed_to_change_own_feedback(self, serializer):
        assignment = serializer.validated_data['assignment_pk']
        youngest = models.TutorSubmissionAssignment.objects\
            .filter(submission=assignment.submission)\
            .order_by('-created')\
            .first()

        return assignment == youngest

    def _tutor_attempts_to_patch_first_feedback_final(self, serializer):
        is_final_set = serializer.validated_data.get('is_final', False)
        assignment = serializer.validated_data.get('assignment_pk')
        in_creation = assignment.subscription.feedback_stage == models.GeneralTaskSubscription.FEEDBACK_CREATION  # noqa
        return is_final_set and in_creation

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        if self._tutor_attempts_to_set_first_feedback_final(serializer):
            return Response(
                {'It is not allowed to create feedback final for tutors'},
                status=status.HTTP_403_FORBIDDEN)

        if self._request_user_does_not_own_assignment(serializer):
            return Response(
                {'This user has not permission to create this feedback'},
                status=status.HTTP_403_FORBIDDEN)

        self.perform_create(serializer)
        return Response(serializer.data,
                        status=status.HTTP_201_CREATED)

    def partial_update(self, request, **kwargs):
        feedback = self.get_object()
        assignment = models.TutorSubmissionAssignment.objects.get(
            subscription__owner=request.user,
            submission=feedback.of_submission
        )
        serializer = self.get_serializer(
            feedback,
            data={'assignment_pk': assignment.pk,
                  'score': feedback.score,
                  **request.data},
            partial=True)

        serializer.is_valid(raise_exception=True)

        if self._tutor_attempts_to_change_final_feedback(serializer):
            return Response(
                {'Cannot set the first feedback final unless user reviewer'},
                status=status.HTTP_403_FORBIDDEN)

        if self._tutor_attempts_to_patch_first_feedback_final(serializer):
            return Response(
                {'Cannot set the first feedback final unless user reviewer'},
                status=status.HTTP_403_FORBIDDEN)

        if not self._tutor_is_allowed_to_change_own_feedback(serializer):
            return Response(
                {'Sorry, but somebody else is already working on this on'},
                status=status.HTTP_403_FORBIDDEN)

        serializer.save()
        return Response(serializer.data)
