import logging

from rest_framework import mixins, status, viewsets
from rest_framework.decorators import detail_route
from rest_framework.response import Response

from core import models, permissions, serializers
from core.models import TutorSubmissionAssignment
from core.permissions import IsTutorOrReviewer
from core.serializers import AssignmentSerializer

log = logging.getLogger(__name__)


class SubscriptionApiViewSet(
        mixins.RetrieveModelMixin,
        mixins.CreateModelMixin,
        mixins.ListModelMixin,
        viewsets.GenericViewSet):
    permission_classes = (permissions.IsTutorOrReviewer,)
    queryset = models.GeneralTaskSubscription.objects.all()
    serializer_class = serializers.SubscriptionSerializer

    def fetch_assignment(self,
                         request,
                         assignment_fetcher,
                         error_message):
        try:
            assignment = assignment_fetcher()
        except models.SubscriptionEnded as err:
            return Response(
                {'Error': error_message},
                status=status.HTTP_410_GONE)
        serializer = serializers.AssignmentDetailSerializer(assignment)
        return Response(serializer.data)

    @detail_route(methods=['get'], url_path='assignments/current')
    def current_assignment(self, request, pk=None):
        subscription = self.get_object()

        return self.fetch_assignment(
            request,
            subscription.get_oldest_unfinished_assignment,
            'Seems this subscription has ended')

    @detail_route(methods=['get'], url_path='assignments/next')
    def next_assignment(self, request, pk=None):
        subscription = self.get_object()

        return self.fetch_assignment(
            request,
            subscription.get_youngest_unfinished_assignment,
            'Seems there is nothing left to prefetch')

    def get_queryset(self):
        return models.GeneralTaskSubscription.objects.filter(
            owner=self.request.user)

    def destroy(self, request, pk=None):
        instance = self.get_object()

        raise NotImplemented  # todo: just impl. deactivate and reactivate

        instance.delete()
        return Response(status=status.HTTP_204_NO_CONTENT)

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        subscription = serializer.save()

        if subscription.query_type == models.GeneralTaskSubscription.STUDENT_QUERY:  # noqa TODO refactor name of generaltasksubscription to SubmissionSubscription
            subscription.reserve_all_assignments_for_a_student()
        else:
            subscription.get_oldest_unfinished_assignment()

        headers = self.get_success_headers(serializer.data)
        return Response(serializer.data,
                        status=status.HTTP_201_CREATED,
                        headers=headers)


class AssignmentApiViewSet(
        mixins.RetrieveModelMixin,
        mixins.ListModelMixin,
        viewsets.GenericViewSet):
    permission_classes = (IsTutorOrReviewer,)
    queryset = TutorSubmissionAssignment.objects.all()
    serializer_class = AssignmentSerializer

    def get_queryset(self):
        """ Get only assignments of that user """
        return TutorSubmissionAssignment.objects.filter(
            subscription__owner=self.request.user)

    def destroy(self, request, pk=None):
        """ Stop working on the assignment before it is finished """
        instance = self.get_object()

        if instance.is_done:
            return Response(status=status.HTTP_403_FORBIDDEN)  # test

        instance.delete()
        return Response(status=status.HTTP_204_NO_CONTENT)  # test
